import { DRAFT_STATE } from './env.js';
import { ProxyType } from './types-internal.js';
import './common.js';
import './finalize.js';
import './proxy.js';
import './immerClass.js';
import './plugins.js';

/** Each scope represents a `produce` call. */

var ImmerScope =
/*#__PURE__*/
function () {
  function ImmerScope(parent, immer) {
    this.drafts = [];
    this.parent = parent;
    this.immer = immer; // Whenever the modified draft contains a draft from another scope, we
    // need to prevent auto-freezing so the unowned draft can be finalized.

    this.canAutoFreeze = true;
  }

  var _proto = ImmerScope.prototype;

  _proto.usePatches = function usePatches(patchListener) {
    if (patchListener) {
      this.patches = [];
      this.inversePatches = [];
      this.patchListener = patchListener;
    }
  };

  _proto.revoke = function revoke() {
    this.leave();
    this.drafts.forEach(_revoke); // @ts-ignore

    this.drafts = null;
  };

  _proto.leave = function leave() {
    if (this === ImmerScope.current) {
      ImmerScope.current = this.parent;
    }
  };

  ImmerScope.enter = function enter(immer) {
    var scope = new ImmerScope(ImmerScope.current, immer);
    ImmerScope.current = scope;
    return scope;
  };

  return ImmerScope;
}();

function _revoke(draft) {
  var state = draft[DRAFT_STATE];
  if (state.type === ProxyType.ProxyObject || state.type === ProxyType.ProxyArray) state.revoke();else state.revoked = true;
}

export { ImmerScope };
